// main.cu

#include <stdio.h>
#include <conio.h>
#include <malloc.h>
#include <windows.h>

#define TRUE 1
#define FALSE 0

#define NOT_FULL_COLUMN -1
#define MAX_CELL_VALUE 1

#define EMPTY_CELL 3
#define NO_CANDIDS 2

#define MAXLEVEL 63

#define DEFAULT_PAR_LEVEL 15

#define CUBES_PER_RUN 200
#define CUBES_PER_BLOCK 24

#define BLOCKORDER 10
#define THREADORDER 10

const int	BlockNum = 1 << BLOCKORDER,
			ThreadNum = 1 << THREADORDER;

#define EVENT_BREAK 1
#define EVENT_NONE 0

#define RUN_SAVE_SPAN 100

#define CUDA_SUCCESS 0
#define CUDA_ERR_TIMEOUT 1

const char	resultFile[] = "store.txt",
			persistentFile[] = "save.hxc";

typedef int cubeCells[4][4][4];

typedef struct {
	cubeCells cells;
	int currSum;
	int numX[16];
	int numY[16];
	int numZ[16];
	int level;
} cube;

typedef struct {
	int cubeCount;
	int parRuns;
	int maxCubesInRun;
	int maxCubesInThread;
} cubeStats;

typedef struct {
	int cubeCount;
	int maxCubesInThread;
} parStats;

void restoreOrInitialize(cube* pcurvar, cubeStats* pstats);
void restoreFromFile(FILE* pfSave, cube* pcurvar, cubeStats* pstats);
void initializeCube(cube* pcurvar, cubeStats* pstats);
void showStartStatistics(const cube* pcurvar, const cubeStats* pstats);
void showEndStatistics(const cubeStats* pstats);
void saveCube(const cube* pcurvar, const cubeStats* pstats);

void iterateCubes(cube* pcurvar, cubeStats* pstats, int parLevel);

__host__ __device__ inline int getLevel(const cube* pcurvar);
__host__ __device__ inline int calcX(int level);
__host__ __device__ inline int calcY(int level);
__host__ __device__ inline int calcZ(int level);
__host__ __device__ int determineCandid(const cube* pcurvar, int prevCand);
__host__ __device__ BOOL checkLevelFill(const cube* pcurvar);
__host__ __device__ BOOL checkLevelFillWithValue(const cube* pcurvar, int val);
__host__ __device__ BOOL maxValueReached(const cube* pcurvar, int candid);
__host__ __device__ void fillLevel(cube* pcurvar);
__host__ __device__ void fillLevelWithValue(cube* pcurvar, int val);
__host__ __device__ void levelUp(cube* pcurvar);
__host__ __device__ void levelDown(cube* pcurvar);
__host__ __device__ void decrementNums(cube* pcurvar);
__host__ __device__ inline int getLevelValue(const cube* pcurvar);
__host__ __device__ inline int getExplLevelValue(const cube* pcurvar, int level);
__host__ __device__ void setLevelValue(cube* pcurvar, int val);
__host__ __device__ void getColumnX(const cube* pcurvar, int cy, int cz, int* rezt);
__host__ __device__ void getColumnY(const cube* pcurvar, int cx, int cz, int* rezt);
__host__ __device__ void getColumnZ(const cube* pcurvar, int cx, int cy, int* rezt);
__host__ __device__ inline int calcNumber(const int* valcol);
__host__ __device__ void copyCells(const cube* pcurvar, cubeCells& dstCells);
__host__ __device__ void copyCube(cube* cubeDst, const cube* cubeSrc);

void saveCells(const cubeCells* pcells, int count);
int getNumberString(const cube* pcurvar, int parlevel);
int getEvent();

#ifdef USING_CUDA
typedef struct {
	int* parCount;
	int* reztCount;
	cubeCells* parStore;
	cube* parvar;
} memAllocs;

int runParallelCudaIter(const cube* pcurvar, parStats* pparstats, cubeCells* resultCells, memAllocs& workmem);
__global__ void iterateCubesCuda(const cube* pmainvar, int* count, cubeCells* rezt);
#else
void runParallelIterating(const cube* pcurvar, parStats* pparstats, cubeCells* resultCells);
void iterateThread(const cube* pmainvar, int threadid, int* count, cubeCells* rezt);
#endif

int main(int argc, char** argv) {
	cube* pCurvar;
	pCurvar = (cube*) malloc(sizeof(cube));
	cubeStats* pIterStats;
	pIterStats = (cubeStats*) malloc(sizeof(cubeStats));

	restoreOrInitialize(pCurvar, pIterStats);
	showStartStatistics(pCurvar, pIterStats);
	iterateCubes(pCurvar, pIterStats, DEFAULT_PAR_LEVEL);
	saveCube(pCurvar, pIterStats);
	showEndStatistics(pIterStats);

	free(pIterStats);
	free(pCurvar);
	_getch();
	return 0;
}

void restoreOrInitialize(cube* pcurvar, cubeStats* pstats) {
	FILE *pfSave;
	int errcode = fopen_s(&pfSave, persistentFile, "rt");
	if(pfSave != NULL) {
		restoreFromFile(pfSave, pcurvar, pstats);
		fclose(pfSave);
	} else
		initializeCube(pcurvar, pstats);
	return;
}

void restoreFromFile(FILE* pfSave, cube* pcurvar, cubeStats* pstats) {
	fscanf_s(pfSave, "%d", &pcurvar->level);
	fscanf_s(pfSave, "%d", &pstats->parRuns);
	fscanf_s(pfSave, "%d", &pstats->cubeCount);
	fscanf_s(pfSave, "%d", &pstats->maxCubesInRun);
	fscanf_s(pfSave, "%d", &pstats->maxCubesInThread);
	fscanf_s(pfSave, "%d", &pcurvar->currSum);

	for(int w = 0; w < 4; w++) {
		for(int v = 0; v < 4; v++) {
			for(int u = 0; u < 4; u++) {
				fscanf_s(pfSave, "%d", &pcurvar->cells[w][v][u]);
			}
		}
	}
	int* pNumRefs[3]; pNumRefs[0] = pcurvar->numX; pNumRefs[1] = pcurvar->numY; pNumRefs[2] = pcurvar->numZ;
	for(int i = 0; i < 3; i++) {
		for(int j = 0; j < 16; j++) {
			fscanf_s(pfSave, "%d", &pNumRefs[i][j]);
		}
	}
	return;
}

void initializeCube(cube* pcurvar, cubeStats* pstats) {
	pcurvar->level = 0;
	pcurvar->currSum = 0;

	for(int w = 0; w < 4; w++) {
		for(int v = 0; v < 4; v++) {
			for(int u = 0; u < 4; u++) {
				pcurvar->cells[w][v][u] = EMPTY_CELL;
			}
		}
	}
	int* pNumRefs[3]; pNumRefs[0] = pcurvar->numX; pNumRefs[1] = pcurvar->numY; pNumRefs[2] = pcurvar->numZ;
	for(int i = 0; i < 3; i++) {
		for(int j = 0; j < 16; j++) {
			pNumRefs[i][j] = 0;
		}
	}

	pstats->cubeCount = 0;
	pstats->maxCubesInRun = 0;
	pstats->maxCubesInThread = 0;
	pstats->parRuns = 0;
	return;
}

void showStartStatistics(const cube* pcurvar, const cubeStats* pstats) {
	printf("Stop level: %d, cubes found: %d, runs: %d\n", pcurvar->level, pstats->cubeCount, pstats->parRuns);
	return;
}

void showEndStatistics(const cubeStats* pstats) {
	printf("Cubes found so far: %d, max cubes per run: %d, per thread: %d\n",
				pstats->cubeCount, pstats->maxCubesInRun, pstats->maxCubesInThread);
	return;
}

void saveCube(const cube* pcurvar, const cubeStats* pstats) {
	FILE* pfStore;
	int errcode = fopen_s(&pfStore, persistentFile, "wt");

	fprintf(pfStore, "%d\n", pcurvar->level);
	fprintf(pfStore, "%d\n", pstats->parRuns);
	fprintf(pfStore, "%d\n", pstats->cubeCount);
	fprintf(pfStore, "%d\n", pstats->maxCubesInRun);
	fprintf(pfStore, "%d\n", pstats->maxCubesInThread);
	fprintf(pfStore, "%d\n", pcurvar->currSum);

	for(int w = 0; w < 4; w++) {
		for(int v = 0; v < 4; v++) {
			for(int u = 0; u < 4; u++) {
				fprintf(pfStore, "%d\n", pcurvar->cells[w][v][u]);
			}
		}
	}
	const int* pNumRefs[3]; pNumRefs[0] = pcurvar->numX; pNumRefs[1] = pcurvar->numY; pNumRefs[2] = pcurvar->numZ;
	for(int i = 0; i < 3; i++) {
		for(int j = 0; j < 16; j++) {
			fprintf(pfStore, "%d\n", pNumRefs[i][j]);
		}
	}
	fclose(pfStore);
	return;
}

__host__ __device__ inline int getLevel(const cube* pcurvar) {
	return pcurvar->level;
}

__host__ __device__ inline int calcX(int level) {
	return level % 4;
}

__host__ __device__ inline int calcY(int level) {
	return (level / 4) % 4;
}

__host__ __device__ inline int calcZ(int level) {
	return level / 16;
}

__host__ __device__ int determineCandid(const cube* pcurvar, int prevCand) {
	int curval = getLevelValue(pcurvar);
	if(curval == 0)
		return 1;
	if(curval == 1 || prevCand == 1)
		return NO_CANDIDS;
	if(prevCand == EMPTY_CELL)
		return 0;
	return 1;
}

__host__ __device__ BOOL checkLevelFillWithValue(const cube* pcurvar, int val) {
	if(val == NO_CANDIDS)
		return FALSE;

	int colX[4], colY[4], colZ[4];
	getColumnX(pcurvar, calcY(pcurvar->level), calcZ(pcurvar->level), colX);
	colX[calcX(pcurvar->level)] = val;
	int nmx = calcNumber(colX);
	if(nmx != NOT_FULL_COLUMN) {
		if(pcurvar->numX[nmx] > 0)
			return FALSE;
	}
	getColumnY(pcurvar, calcX(pcurvar->level), calcZ(pcurvar->level), colY);
	colY[calcY(pcurvar->level)] = val;
	int nmy = calcNumber(colY);
	if(nmy != NOT_FULL_COLUMN) {
		if(pcurvar->numY[nmy] > 0)
			return FALSE;
	}
	getColumnZ(pcurvar, calcX(pcurvar->level), calcY(pcurvar->level), colZ);
	colZ[calcZ(pcurvar->level)] = val;
	int nmz = calcNumber(colZ);
	if(nmz != NOT_FULL_COLUMN) {
		if(pcurvar->numZ[nmz] > 0)
			return FALSE;
	}
	return TRUE;
}

__host__ __device__ void fillLevelWithValue(cube* pcurvar, int val) {
	if(getLevelValue(pcurvar) != EMPTY_CELL)
		decrementNums(pcurvar);

	setLevelValue(pcurvar, val);
	pcurvar->currSum += val;

	int colX[4], colY[4], colZ[4];

	getColumnX(pcurvar, calcY(pcurvar->level), calcZ(pcurvar->level), colX);
	int nmx = calcNumber(colX);
	if(nmx != NOT_FULL_COLUMN) {
		pcurvar->numX[nmx]++;
	}
	getColumnY(pcurvar, calcX(pcurvar->level), calcZ(pcurvar->level), colY);
	int nmy = calcNumber(colY);
	if(nmy != NOT_FULL_COLUMN) {
		pcurvar->numY[nmy]++;
	}
	getColumnZ(pcurvar, calcX(pcurvar->level), calcY(pcurvar->level), colZ);
	int nmz = calcNumber(colZ);
	if(nmz != NOT_FULL_COLUMN) {
		pcurvar->numZ[nmz]++;
	}
	return;
}

__host__ __device__ BOOL maxValueReached(const cube* pcurvar, int candid) {
	if(getLevelValue(pcurvar) == 1)
		return TRUE;
	if(candid == 1)
		return TRUE;
	return FALSE;
}

__host__ __device__ void levelUp(cube* pcurvar) {
	if(pcurvar->level < MAXLEVEL)
		pcurvar->level++;
}

__host__ __device__ void levelDown(cube* pcurvar) {
	if(getLevelValue(pcurvar) != EMPTY_CELL) {
		decrementNums(pcurvar);
		pcurvar->currSum -= getLevelValue(pcurvar);
		setLevelValue(pcurvar, EMPTY_CELL);
	}
	pcurvar->level--;
	return;
}

__host__ __device__ void decrementNums(cube* pcurvar) {
	int colX[4], colY[4], colZ[4];

	getColumnX(pcurvar, calcY(pcurvar->level), calcZ(pcurvar->level), colX);
	int nmx = calcNumber(colX);
	if(nmx != NOT_FULL_COLUMN) {
		pcurvar->numX[nmx]--;
	}
	getColumnY(pcurvar, calcX(pcurvar->level), calcZ(pcurvar->level), colY);
	int nmy = calcNumber(colY);
	if(nmy != NOT_FULL_COLUMN) {
		pcurvar->numY[nmy]--;
	}
	getColumnZ(pcurvar, calcX(pcurvar->level), calcY(pcurvar->level), colZ);
	int nmz = calcNumber(colZ);
	if(nmz != NOT_FULL_COLUMN) {
		pcurvar->numZ[nmz]--;
	}
	return;
}

__host__ __device__ inline int getLevelValue(const cube* pcurvar) {
	return pcurvar->cells[calcZ(pcurvar->level)][calcY(pcurvar->level)][calcX(pcurvar->level)];
}

__host__ __device__ inline int getExplLevelValue(const cube* pcurvar, int level) {
	return pcurvar->cells[calcZ(level)][calcY(level)][calcX(level)];
}

__host__ __device__ void setLevelValue(cube* pcurvar, int val) {
	pcurvar->cells[calcZ(pcurvar->level)][calcY(pcurvar->level)][calcX(pcurvar->level)] = val;
	return;
}

__host__ __device__ void getColumnX(const cube* pcurvar, int cy, int cz, int* rezt) {
	rezt[0] = pcurvar->cells[cz][cy][0];
	rezt[1] = pcurvar->cells[cz][cy][1];
	rezt[2] = pcurvar->cells[cz][cy][2];
	rezt[3] = pcurvar->cells[cz][cy][3];
	return;
}

__host__ __device__ void getColumnY(const cube* pcurvar, int cx, int cz, int* rezt) {
	rezt[0] = pcurvar->cells[cz][0][cx];
	rezt[1] = pcurvar->cells[cz][1][cx];
	rezt[2] = pcurvar->cells[cz][2][cx];
	rezt[3] = pcurvar->cells[cz][3][cx];
	return;
}

__host__ __device__ void getColumnZ(const cube* pcurvar, int cx, int cy, int* rezt) {
	rezt[0] = pcurvar->cells[0][cy][cx];
	rezt[1] = pcurvar->cells[1][cy][cx];
	rezt[2] = pcurvar->cells[2][cy][cx];
	rezt[3] = pcurvar->cells[3][cy][cx];
	return;
}

__host__ __device__ inline int calcNumber(const int* valcol) {
	if(valcol[0] == EMPTY_CELL || valcol[1] == EMPTY_CELL ||
		valcol[2] == EMPTY_CELL || valcol[3] == EMPTY_CELL)
		return NOT_FULL_COLUMN;

	return (valcol[3] << 3) + (valcol[2] << 2) + (valcol[1] << 1) + valcol[0];
}

__host__ __device__ void copyCells(const cube* pcurvar, cubeCells& dstCells) {
	for(int w = 0; w < 4; w++) {
		for(int v = 0; v < 4; v++) {
			for(int u = 0; u < 4; u++) {
				dstCells[w][v][u] = pcurvar->cells[w][v][u];
			}
		}
	}
	return;
}

__host__ __device__ void copyCube(cube* cubeDst, const cube* cubeSrc) {
	for(int w = 0; w < 4; w++) {
		for(int v = 0; v < 4; v++) {
			for(int u = 0; u < 4; u++) {
				cubeDst->cells[w][v][u] = cubeSrc->cells[w][v][u];
			}
		}
	}
	const int* pSrcRefs[3]; pSrcRefs[0] = cubeSrc->numX; pSrcRefs[1] = cubeSrc->numY; pSrcRefs[2] = cubeSrc->numZ;
	int* pDstRefs[3]; pDstRefs[0] = cubeDst->numX; pDstRefs[1] = cubeDst->numY; pDstRefs[2] = cubeDst->numZ;
	for(int i = 0; i < 3; i++) {
		for(int j = 0; j < 16; j++) {
			pDstRefs[i][j] = pSrcRefs[i][j];
		}
	}
	cubeDst->level = cubeSrc->level;
	cubeDst->currSum = cubeSrc->currSum;
	return;
}

void saveCells(const cubeCells* pcells, int count) {
	FILE *pfCubes;
	int errcode;
	do {
		errcode = fopen_s(&pfCubes, resultFile, "a+t");
	} while(errcode != 0);

	for(int i = 0; i < count; i++) {
		for(int w = 0; w < 4; w++) {
			for(int v = 0; v < 4; v++) {
				for(int u = 0; u < 4; u++) {
					fprintf(pfCubes, "%d", pcells[i][w][v][u]);
				}
			}
		}
		fprintf(pfCubes, "\n");
	}
	fclose(pfCubes);
	return;
}

int getNumberString(const cube* pcurvar, int parlevel) {
	int num = 0;
	for(int i = parlevel - 1, j = 0; i >= 0; i--, j++) {
		num += getExplLevelValue(pcurvar, i) << j;
	}
	return num;
}

int getEvent() {
	if(_kbhit() == 0)
		return EVENT_NONE;

	int kc = _getch_nolock();
	if(kc == 'e' || kc == 'E')
		return EVENT_BREAK;
	return EVENT_NONE;
}

#ifdef USING_CUDA   //////////////////////////////////// CUDA block /////////////////////////////////////////////////////////
BOOL ifCudaError(const char* srcmsg) {
	cudaError_t cudaErr = cudaGetLastError();
	if(cudaErr != cudaSuccess) {
		if(cudaErr == cudaErrorLaunchTimeout) {
			printf("Timeout (%s)\n", srcmsg);
			return TRUE;
		}
		printf("%s: %s (%u)\n", srcmsg, cudaGetErrorString(cudaErr), cudaErr);
		_getch();
		exit(-1);
	}
	return FALSE;
}

void prepareCubeMemory(memAllocs& mem) {
	mem.reztCount = (int*) malloc(sizeof(int) * BlockNum);

	while(TRUE) {
		cudaMalloc((void**) &mem.parvar, sizeof(cube));
		if(ifCudaError("Malloc cube") == FALSE)
			break;
		cudaFree((void*) mem.parvar);
	}

	while(TRUE) {
		cudaMalloc((void**) &mem.parCount, sizeof(int) * BlockNum);
		if(ifCudaError("Malloc device counts") == FALSE)
			break;
		cudaFree((void*) mem.parCount);
	}

	while(TRUE) {
		cudaMalloc((void**) &mem.parStore, sizeof(cubeCells) * BlockNum * CUBES_PER_BLOCK);
		if(ifCudaError("Malloc block results") == FALSE)
			break;
		cudaFree((void*) mem.parStore);
	}
	return;
}

void freeCubeMemory(memAllocs& mem) {
	free(mem.reztCount);
	while(TRUE) {
		cudaFree(mem.parvar);
		// if(ifCudaError("Free cube") == FALSE)
			break;
	}
	while(TRUE) {
		cudaFree(mem.parCount);
		// if(ifCudaError("Free device counts") == FALSE)
			break;
	}
	while(TRUE) {
		cudaFree(mem.parStore);
		// if(ifCudaError("Free block results") == FALSE)
			break;
	}
	return;
}

void iterateCubes(cube* pcurvar, cubeStats* pstats, int parLevel) {
	cubeCells resultCells[CUBES_PER_RUN];
	int candid = EMPTY_CELL;
	memAllocs cudamem;
	prepareCubeMemory(cudamem);
	int currParLevel = parLevel;

	while(getLevel(pcurvar) >= 0) {
		if(currParLevel != parLevel) {
			if(getLevel(pcurvar) == parLevel)
				currParLevel = parLevel;
		}

		if(getLevel(pcurvar) == currParLevel) {
			parStats runStat; runStat.cubeCount = 0;
			DWORD parTimeStart = GetTickCount();
			int iterError = runParallelCudaIter(pcurvar, &runStat, resultCells, cudamem);
			if(iterError == CUDA_ERR_TIMEOUT) {
				currParLevel++;
				printf("Timeout, increasing level to %d\n", currParLevel);
				freeCubeMemory(cudamem);
				while(TRUE) {
					cudaDeviceReset();
					if(ifCudaError("Resetting device") == FALSE)
						break;
				}
				prepareCubeMemory(cudamem);
			} else {
				DWORD parTimeEnd = GetTickCount();
				pstats->parRuns++;
				if(runStat.cubeCount > 0) {
					saveCells(resultCells, runStat.cubeCount);
					pstats->cubeCount += runStat.cubeCount;
					if(pstats->maxCubesInRun < runStat.cubeCount)
						pstats->maxCubesInRun = runStat.cubeCount;
					if(currParLevel == parLevel)
						saveCube(pcurvar, pstats);
				} else if(pstats->parRuns % RUN_SAVE_SPAN == 0) {
					if(currParLevel == parLevel)
						saveCube(pcurvar, pstats);
				}
				printf("Got: %d, total: %d, ms: %u,\trun: %d, num: %d\n",
						runStat.cubeCount, pstats->cubeCount, parTimeEnd - parTimeStart,
						pstats->parRuns, getNumberString(pcurvar, currParLevel));

				if(currParLevel == parLevel) {
					int userEvent = getEvent();
					if(userEvent == EVENT_BREAK)
						break;
				}

				levelDown(pcurvar);
				candid = EMPTY_CELL;
			}
		}
		candid = determineCandid(pcurvar, candid);
		if(checkLevelFillWithValue(pcurvar, candid) == TRUE) {
			fillLevelWithValue(pcurvar, candid);
			if(getLevel(pcurvar) < currParLevel) {
				levelUp(pcurvar);
				candid = EMPTY_CELL;
			}
		} else {
			if(maxValueReached(pcurvar, candid) == TRUE) {
				levelDown(pcurvar);
				candid = EMPTY_CELL;
			}
		}
	}
	return;
}

int runParallelCudaIter(const cube* pcurvar, parStats* pparstats, cubeCells* resultCells, memAllocs& workmem) {
	while(TRUE) {
		cudaMemcpy(workmem.parvar, pcurvar, sizeof(cube), cudaMemcpyHostToDevice);
		if(ifCudaError("Copy cube to device") == FALSE)
			break;
	}

	iterateCubesCuda<<<BlockNum, ThreadNum>>>(workmem.parvar, workmem.parCount, workmem.parStore);
	if(ifCudaError("Parallel iterating") == TRUE)
		return CUDA_ERR_TIMEOUT;

	cudaMemcpy(workmem.reztCount, workmem.parCount, sizeof(int) * BlockNum, cudaMemcpyDeviceToHost);
	if(ifCudaError("Copying counts to host") == TRUE)
		return CUDA_ERR_TIMEOUT;

	int totalCount = 0;
	for(int i = 0; i < BlockNum; i++) {
		totalCount += workmem.reztCount[i];
	}

	if(totalCount > 0) {
		int pos = 0;
		for(int i = 0; i < BlockNum; i++) {
			if(workmem.reztCount[i] > 0) {
				while(TRUE) {
					cudaMemcpy(&resultCells[pos], &workmem.parStore[CUBES_PER_BLOCK * i],
									sizeof(cubeCells) * workmem.reztCount[i], cudaMemcpyDeviceToHost);
					if(ifCudaError("Copying cubes found") == FALSE)
						break;
				}
				pos += workmem.reztCount[i];
			}
		}
	}
	pparstats->cubeCount = totalCount;
	return CUDA_SUCCESS;
}

__constant__ int LevelsFill = BLOCKORDER + THREADORDER;

__global__ void iterateCubesCuda(const cube* pmainvar, int* count, cubeCells* rezt) {
	__shared__ unsigned int cubeIdx;
	if(threadIdx.x == 0) {
		cubeIdx = 0;
	}
	__syncthreads();

	cube localvar;
	copyCube(&localvar, pmainvar);

	int fullvarid = (blockIdx.x << THREADORDER) + threadIdx.x;
	BOOL goodFill = TRUE;

	for(int i = LevelsFill - 1; i >= 0; i--) {
		int mask = 1 << i;
		int candid = ((fullvarid & mask) == 0 ? 0 : 1);
		if(checkLevelFillWithValue(&localvar, candid) == FALSE) {
			goodFill = FALSE;
			break;
		}
		fillLevelWithValue(&localvar, candid);
		levelUp(&localvar);
	}

	if(goodFill == TRUE) {
		int limitlevel = getLevel(&localvar);
		int candid = EMPTY_CELL;
		int cubeFilled = FALSE;

		while(getLevel(&localvar) >= limitlevel) {
			if(getLevel(&localvar) == MAXLEVEL && cubeFilled == TRUE) {
				unsigned int myidx = atomicInc(&cubeIdx, CUBES_PER_BLOCK);
				copyCells(&localvar, rezt[CUBES_PER_BLOCK * blockIdx.x + myidx]);
			}
			candid = determineCandid(&localvar, candid);
			if(checkLevelFillWithValue(&localvar, candid) == TRUE) {
				fillLevelWithValue(&localvar, candid);
				if(getLevel(&localvar) < MAXLEVEL) {
					levelUp(&localvar);
					candid = EMPTY_CELL;
				} else {
					cubeFilled = TRUE;
				}
			} else {
				if(maxValueReached(&localvar, candid) == TRUE) {
					levelDown(&localvar);
					candid = EMPTY_CELL;
					cubeFilled = FALSE;
				}
			}
		}
	}
	__syncthreads();

	if(threadIdx.x == 0) {
		count[blockIdx.x] = (int) cubeIdx;
	}
	return;
}
#else ////////////////////////////////////////////////// non-CUDA block ////////////////////////////////////////////////
void iterateCubes(cube* pcurvar, cubeStats* pstats, int parLevel) {
	cubeCells resultCells[CUBES_PER_RUN];
	int candid = EMPTY_CELL;

	while(getLevel(pcurvar) >= 0) {
		if(getLevel(pcurvar) == parLevel) {
			parStats runStat; runStat.cubeCount = 0; runStat.maxCubesInThread = 0;
			DWORD parTimeStart = GetTickCount();
			runParallelIterating(pcurvar, &runStat, resultCells);
			DWORD parTimeEnd = GetTickCount();
			pstats->parRuns++;
			if(runStat.cubeCount > 0) {
				saveCells(resultCells, runStat.cubeCount);
				pstats->cubeCount += runStat.cubeCount;
				if(pstats->maxCubesInThread < runStat.maxCubesInThread)
					pstats->maxCubesInThread = runStat.maxCubesInThread;
				if(pstats->maxCubesInRun < runStat.cubeCount)
					pstats->maxCubesInRun = runStat.cubeCount;
				saveCube(pcurvar, pstats);
			} else if(pstats->parRuns % RUN_SAVE_SPAN == 0) {
				saveCube(pcurvar, pstats);
			}
			printf("Got: %d (%d), total: %d, ms: %u,\trun: %d, num: %d\n",
					runStat.cubeCount, runStat.maxCubesInThread, pstats->cubeCount, parTimeEnd - parTimeStart,
					pstats->parRuns, getNumberString(pcurvar, parLevel));
			int userEvent = getEvent();
			if(userEvent == EVENT_BREAK)
				break;

			levelDown(pcurvar);
			candid = EMPTY_CELL;
		}
		candid = determineCandid(pcurvar, candid);
		if(checkLevelFillWithValue(pcurvar, candid) == TRUE) {
			fillLevelWithValue(pcurvar, candid);
			if(getLevel(pcurvar) < parLevel) {
				levelUp(pcurvar);
				candid = EMPTY_CELL;
			}
		} else {
			if(maxValueReached(pcurvar, candid) == TRUE) {
				levelDown(pcurvar);
				candid = EMPTY_CELL;
			}
		}
	}
	return;
}

void runParallelIterating(const cube* pcurvar, parStats* pparstats, cubeCells* resultCells) {
	pparstats->cubeCount = 0;
	pparstats->maxCubesInThread = 0;
	int thdCubeCnt;

	for(int i = 0; i < ThreadNum; i++) {
		thdCubeCnt = 0;
		iterateThread(pcurvar, i, &thdCubeCnt, &resultCells[pparstats->cubeCount]);
		if(thdCubeCnt > 0) {
			pparstats->cubeCount += thdCubeCnt;
			if(pparstats->maxCubesInThread < thdCubeCnt)
				pparstats->maxCubesInThread = thdCubeCnt;
		}
	}
	return;
}

void iterateThread(const cube* pmainvar, int threadid, int* count, cubeCells* rezt) {
	*count = 0;
	cube localvar;
	copyCube(&localvar, pmainvar);
	for(int i = THREADORDER - 1; i >= 0; i--) {
		int mask = 1 << i;
		int val = ((threadid & mask) == 0 ? 0 : 1);
		if(checkLevelFillWithValue(&localvar, val) == FALSE)
			return;
		fillLevelWithValue(&localvar, val);
		levelUp(&localvar);
	}
	int lowlimit = getLevel(&localvar);
	int candid = EMPTY_CELL;
	BOOL cubeFilled = FALSE;
	while(getLevel(&localvar) >= lowlimit) {
		if(getLevel(&localvar) == MAXLEVEL && cubeFilled == TRUE) {
			copyCells(&localvar, rezt[*count]);
			(*count)++;
		}
		candid = determineCandid(&localvar, candid);
		if(checkLevelFillWithValue(&localvar, candid) == TRUE) {
			fillLevelWithValue(&localvar, candid);
			if(getLevel(&localvar) < MAXLEVEL) {
				levelUp(&localvar);
				candid = EMPTY_CELL;
			} else {
				cubeFilled = TRUE;
			}
		} else {
			if(maxValueReached(&localvar, candid) == TRUE) {
				levelDown(&localvar);
				candid = EMPTY_CELL;
				cubeFilled = FALSE;
			}
		}
	}
	return;
}
#endif
